#first_order_solver
# Demonstration of numeric method techniques to solve a first order differential
# equation of the form dy/dx = f(x,y)
# Euler, Midpoint and Runge-Kutta (fourth order) methods will be employed and contrasted
#
# LAST UPDATED by Andy French December 2017
#
# NOTE ############# means hard coded lines that needs to be modified

#Import Python numeric and plotting libraries
import numpy as Npy
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt

#

####################################

#Define function dy_by_dx = f(x,y)
def f(x,y):
    dybydx = Npy.sin(x*y)
    return dybydx
    
####################################
    
#

#Define Euler method solver
def euler( dx, x0, y0, N ):
    #Initialize output vectors
    x = Npy.zeros((1,N))
    x = x[0]
    x[0] = x0
    y = Npy.zeros((1,N))
    y = y[0]
    y[0] = y0
	
    #Loop through vectors to compute x and y
    for n in range(1,N):
        x[n] = x[n-1] + dx
        y[n] = y[n-1] + dx*f( x[n-1], y[n-1] )
    
    #Set return values of function
    return x, y
    
#
        
#Define Midpoint method solver
def midpoint( dx, x0, y0, N ):
    #Initialize output vectors
    x = Npy.zeros((1,N))
    x = x[0]
    x[0] = x0
    y = Npy.zeros((1,N))
    y = y[0]
    y[0] = y0
	
    #Loop through vectors to compute x and y
    for n in range(1,N):
        x[n] = x[n-1] + dx
        y[n] = y[n-1] + dx*f( x[n-1]  + dx/2, y[n-1] + 0.5*dx*f( x[n-1], y[n-1] ) )
    
    #Set return values of function
    return x, y
    
#
    
#Define Runge-Kutta solver
def rk4( dx, x0, y0, N ):
    #Initialize output vectors
    x = Npy.zeros((1,N))
    x = x[0]
    x[0] = x0
    y = Npy.zeros((1,N))
    y = y[0]
    y[0] = y0
	
    #Loop through vectors to compute x and y
    for n in range(1,N):
        x[n] = x[n-1] + dx
        k1 = f( x[n-1], y[n-1] )
        k2 = f( x[n-1] + dx/2, y[n-1] + 0.5*dx*k1 )
        k3 = f( x[n-1] + dx/2, y[n-1] + 0.5*dx*k2 )
        k4 = f( x[n-1] + dx, y[n-1] + dx*k3 )
        y[n] = y[n-1] + (dx/6)*( k1 + 2*k2 + 2*k3 + k4 )
    
    #Set return values of function
    return x, y

#

#Run simulation

#Message to command window
print("\n First order Ordinary Differential Equation Solver")
print(" By Andy French December 2017 ")
print(" \n Solves the differential equation dy/dx = f(x,y) ")
print(" \n using (i) Euler, (ii) Midpoint and (iii) Runge-Kutta methods")

##############################################################################

#Define x step
dx = 0.1

#Define fontsize
fsize = 18

#Define initial values
x0 = 0
y0 = 1

#Define end x value (approximately)
xmax = 10

#Compute number of data points i.e. xmax = x0 + (N-1)*dx
N = int( (xmax - x0)/dx + 1 )

#Update xmax
xmax = x0 + (N-1)*dx

#Title string
title_str = "Solution of dy/dt = sin(xy), N="+ str(N) + ", dx=" + str(dx)

##############################################################################

#Compute x,y vectors
xe, ye = euler( dx, x0, y0, N )
xmp, ymp = midpoint( dx, x0, y0, N )
xrk4, yrk4 = rk4( dx, x0, y0, N )

#Create figure and axis
fig = plt.figure(figsize=(10, 8))

#Plot y vs x
plt.plot( xe,ye,'-',label = "Euler" , color = [1,0,0] )
plt.plot( xmp,ymp,'-',label = "Midpoint", color = [0,1,0] )
plt.plot( xrk4,yrk4,'-',label = "Runge-Kutta", color = [0,0,1] )
plt.xlabel( "x", size=fsize );
plt.ylabel( "y", size=fsize );
plt.xticks( fontsize = fsize )
plt.yticks( fontsize = fsize )
plt.title( title_str, size=fsize )
plt.legend( fontsize = fsize,loc = "upper right" )
plt.xlim( x0, xmax )
plt.grid(True)

#Export PNG file
fig.savefig('y vs x.png', dpi=300)

#Wait for user input to close terminal window
input("\n Press any key to continue ...")

#End of code
