#mandlebrot
# Python code which computes the Mandlebrot fractal, creating a PNG colour image
# using a MATLAB style colour map.
#
# LAST UPDATED by Andy French May 2018.

#Import required libraries
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import numpy as np

#

#Inputs

#Define size of Argand diagram in pixels
width = 600
height = 400

#Define Argand diagram x,y extent
xmin = -2.5
ymin = -1.2
ymax = 1.2
xmax = xmin + (width/height)*(ymax-ymin)

#Define number of iterations
N = 32

#Define radius of conversion
R = 3

#Map creation rule
map_rule = "Im and Re"

#Define colourmap ("B&W", "duotone","four colour", "jet", "prism")
colour_map_type = "prism"

#

#MATLAB style colour map generating functions
def matlab_colourmaps( colour_map_type, k ):

    if colour_map_type=="four colour":

        cmap_levels = [0,0.2,0.5,0.8]
        cmap = [
            [1,0,0],
            [0,1,0],
            [0,0,1],
            [0,0,0]
               ]

    elif colour_map_type=="duotone":

        cmap = [
            [1,0,0],[0,1,0]
               ]
        cmap_levels = [0,1]

    elif colour_map_type=="B&W":

        cmap = [[0,0,0]]
        cmap_levels = np.linspace(0,1,k)

        for i in range(1,k):
            if (i%2)==0:
                cmap.append( [0,0,0] )
            else:
                cmap.append( [1,1,1] )

    elif colour_map_type=="prism":

        cmap = [
            [1, 0, 0],
            [1,0.5,0],
            [1.0000,1.0000,0],
            [0,1.0000,0],
            [0,0,1.0000],
            [0.6667,0,1.0000],
            [1.0000,0,0],
            [1.0000,0.5000,0],
            [1.0000,1.0000,0],
            [0,1.0000,0],
            [0,0,1.0000],
            [0.6667,0,1.0000],
            [1.0000,0,0],
            [1.0000,0.5000,0],
            [1.0000,1.0000,0],
            [0,1.0000,0],
            [0,0,1.0000],
            [0.6667,0,1.0000],
            [1.0000,0,0],
            [1.0000,0.5000,0],
            [1.0000,1.0000,0],
            [0,1.0000,0],
            [0,0,1.0000],
            [0.6667,0,1.0000],
            [1.0000,0,0],
            [1.0000,0.5000,0],
            [1.0000,1.0000,0],
            [0,1.0000,0],
            [0,0,1.0000],
            [0.6667,0,1.0000],
            [1.0000,0,0],
            [1.0000,0.5000,0],
            [1.0000,1.0000,0],
            [0,1.0000,0],
            [0,0,1.0000],
            [0.6667,0,1.0000],
            [1.0000,0,0],
            [1.0000,0.5000,0],
            [1.0000,1.0000,0],
            [0,1.0000,0],
            [0,0,1.0000],
            [0.6667,0,1.0000],
            [1.0000,0,0],
            [1.0000,0.5000,0],
            [1.0000,1.0000,0],
            [0,1.0000,0],
            [0,0,1.0000],
            [0.6667,0,1.0000],
            [1.0000,0,0],
            [1.0000,0.5000,0],
            [1.0000,1.0000,0],
            [0,1.0000,0],
            [0,0,1.0000],
            [0.6667,0,1.0000],
            [1.0000,0,0],
            [1.0000,0.5000,0],
            [1.0000,1.0000,0],
            [0,1.0000,0],
            [0,0,1.0000],
            [0.6667,0,1.0000],
            [1.0000,0,0],
            [1.0000,0.5000,0],
            [1.0000,1.0000,0],
            [0,1.0000,0]
            ]
        
        #Define linearaly spaced vector [0,1] which corresponds to colour map colours
        cmap_levels = np.linspace(0,1,len(cmap))

    else:
        #MATLAB default "jet" colour map
        cmap = [
            [0,0,0.5625],
            [0,0,0.6250],
            [0,0,0.6875],
            [0,0,0.7500],
            [0,0,0.8125],
            [0,0,0.8750],
            [0,0,0.9375],
            [0,0,1.0000],
            [0,0.0625,1.0000],
            [0,0.1250,1.0000],
            [0,0.1875,1.0000],
            [0,0.2500,1.0000],
            [0,0.3125,1.0000],
            [0,0.3750,1.0000],
            [0,0.4375,1.0000],
            [0,0.5000,1.0000],
            [0,0.5625,1.0000],
            [0,0.6250,1.0000],
            [0,0.6875,1.0000],
            [0,0.7500,1.0000],
            [0,0.8125,1.0000],
            [0,0.8750,1.0000],
            [0,0.9375,1.0000],
            [0,1.0000,1.0000],
            [0.0625,1.0000,0.9375],
            [0.1250,1.0000,0.8750],
            [0.1875,1.0000,0.8125],
            [0.2500,1.0000,0.7500],
            [0.3125,1.0000,0.6875],
            [0.3750,1.0000,0.6250],
            [0.4375,1.0000,0.5625],
            [0.5000,1.0000,0.5000],
            [0.5625,1.0000,0.4375],
            [0.6250,1.0000,0.3750],
            [0.6875,1.0000,0.3125],
            [0.7500,1.0000,0.2500],
            [0.8125,1.0000,0.1875],
            [0.8750,1.0000,0.1250],
            [0.9375,1.0000,0.0625],
            [1.0000,1.0000,0],
            [1.0000,0.9375,0],
            [1.0000,0.8750,0],
            [1.0000,0.8125,0],
            [1.0000,0.7500,0],
            [1.0000,0.6875,0],
            [1.0000,0.6250,0],
            [1.0000,0.5625,0],
            [1.0000,0.5000,0],
            [1.0000,0.4375,0],
            [1.0000,0.3750,0],
            [1.0000,0.3125,0],
            [1.0000,0.2500,0],
            [1.0000,0.1875,0],
            [1.0000,0.1250,0],
            [1.0000,0.0625,0],
            [1.0000,0,0],
            [0.9375,0,0],
            [0.8750,0,0],
            [0.8125,0,0],
            [0.7500,0,0],
            [0.6875,0,0],
            [0.6250,0,0],
            [0.5625,0,0],
            [0.5000,0,0]
            ]

        #Define linearaly spaced vector [0,1] which corresponds to colour map colours
        cmap_levels = np.linspace(0,1,len(cmap))

    #Define vectors for red, green, blue colour map values
    z = np.zeros((1,len(cmap)))[0]
    cmap_red = z.copy()
    cmap_green = z.copy()
    cmap_blue = z.copy()
    for i in range(0,len(cmap)):
        cmap_red[i] = cmap[i][0]
        cmap_green[i] = cmap[i][1]
        cmap_blue[i] = cmap[i][2]
           
    return cmap, cmap_levels, cmap_red, cmap_green, cmap_blue

#

#Display greetings message
print( "\n Python Mandlebrot fractal maker " )
print( " By Dr Andrew French. May 2018 " )

#Request high resolution
mode = input( "\n High resolution mode? (Y/N) \n ")
if mode=="Y":
    width = width*5
    height = height*5

#Initialize colour function C
I = np.ones((height,width))
C = -1*I

#Initialize Argand diagram z0,z
x = np.linspace(xmin,xmax,width)
y = np.linspace(ymin,ymax,height)
x,y = np.meshgrid(x,y)
z0 = x + 1j*y
z = z0.copy()

#Initialize boolean flag to determine whether a point is in or outside
#the Mandlebrot set
M = I.copy()

#Step through N iterations and determine colour function
for n in range(0,N):

    #Mandlebrot transformation
    z[ M==1 ] = np.power( z[ M==1 ], 2*I[ M==1 ] )
    z[ M==1 ] = np.add( z[ M==1 ], z0[ M==1 ] )

    #Define map creation rule
    if map_rule=="Im and Re":
        rule = ( np.abs( np.real(z) ) > R ) & ( np.abs( np.imag(z) ) > R )
    else:
        rule = ( np.abs(z) > R )   

    #Update C and M
    C[ rule & ( M==1 ) ] = n/(N-1)
    M[ rule ] = 0

    #Print a * per iteration to enable user to keep track of progress
    print( " *",end="",flush=True )

#Get colourmap
cmap, cmap_levels, cmap_red, cmap_green, cmap_blue = matlab_colourmaps( colour_map_type, N )

#Form rgb colour array, interpolating colour map using colour function C
print(" \n Interpolating colour map ... ")
rgb = np.zeros((height,width,3))
rgb[:,:,0] = np.interp( C, cmap_levels, cmap_red )
rgb[:,:,1] = np.interp( C, cmap_levels, cmap_green )
rgb[:,:,2] = np.interp( C, cmap_levels, cmap_blue )

#Set C=-1 values to be white
j = 0
for r in range(0,height):
    for c in range(0,width):

        #Compute % done
        j = j + 1
        if 100*j/(height*width) > 10:
            j = 0
            print( " *",end="",flush=True )

        if C[r,c]==-1:
            #White colour if 'within' mandlebrot set
            rgb[r,c,0] = 1
            rgb[r,c,1] = 1
            rgb[r,c,2] = 1
        
#Create figure and axis
print(" \n Plotting RGB image ... ")
fig = plt.figure(figsize=(10, 8))

#Save as PNG bitmap
plt.imsave( "mandlebrot.png", rgb )

#Wait for user input to close terminal window
input("\n Press return to close ")

#End of code




