#second_order_solver
# Demonstration of numeric method techniques to solve a second order differential
# equation of the form dv/dt = a(t,x,v) ;  dx/dt = v
# i.e. Newton's Second Law
#
# A simple Euler and Verlet methods will be contrasted
#
# LAST UPDATED by Andy French December 2017
#
# NOTE ############# means hard coded lines that needs to be modified

#Import Python numeric and plotting libraries
import numpy as Npy
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
pi = float(Npy.pi)

#

####################################

#Define function dv_by_dt = a(t,x,v)
def a(t,x,v):
    P = 1
    w = 2*pi/P
    b = 1
    dvbydt = -(w**2)*x - b*v
    return dvbydt
    
####################################
    
#

#Define Euler method solver
def euler( dt, t0, x0, v0, N ):
    #Initialize output vectors
    t = Npy.zeros((1,N))
    t = t[0]
    t[0] = t0
    x = Npy.zeros((1,N))
    x = x[0]
    x[0] = x0
    v = Npy.zeros((1,N))
    v = v[0]
    v[0] = v0
	
    #Loop through vectors to compute t, x, v
    for n in range(1,N):
        t[n] = t[n-1] + dt
        v[n] = v[n-1] + a( t[n-1], x[n-1], v[n-1] )*dt
        x[n] = x[n-1] + v[n-1]*dt
    
    #Set return values of function
    return t, x, v
    
#
        

    
#Define Verlet solver
def verlet( dt, t0, x0, v0, N ):
    #Initialize output vectors
    t = Npy.zeros((1,N))
    t = t[0]
    t[0] = t0
    x = Npy.zeros((1,N))
    x = x[0]
    x[0] = x0
    v = Npy.zeros((1,N))
    v = v[0]
    v[0] = v0
	
    #Loop through vectors to compute t, x, v
    for n in range(1,N):
        t[n] = t[n-1] + dt
        x[n] = x[n-1] + v[n-1]*dt + 0.5*a( t[n-1], x[n-1], v[n-1] )*(dt**2)
        #Note can't use v[n] in a(n) since we don't know this yet!
        v[n] = v[n-1] + 0.5*( a( t[n-1], x[n-1], v[n-1] ) + a( t[n], x[n], v[n-1] ) )*dt
    
    #Set return values of function
    return t, x, v

#

#Run simulation

#Message to command window
print("\n Second order Ordinary Differential Equation Solver")
print(" By Andy French December 2017 ")
print(" \n Solves the differential equation dv/dt = a(t,x,v) ; dx/dt = v ")
print(" \n using (i) Euler and (ii) Verlet methods")

##############################################################################

#Define t step
dt = 0.01

#Define fontsize
fsize = 18

#Define initial values
t0 = 0
x0 = 1
v0 = 0

#Define end t value (approximately)
tmax = 10

#Set maximum x and v values
xmax = 3
vmax = 10

#Compute number of data points i.e. tmax = t0 + (N-1)*dt
N = int( (tmax - t0)/dt + 1 )

#Update tmax
tmax = t0 + (N-1)*dt

#Title string
title_str = "Solution of dv/dt = -((2*pi/1)^2)*x - 1*v, N="+ str(N) + ", dt=" + str(dt)

##############################################################################

#Compute x,y vectors
te,xe,ve = euler( dt, t0, x0, v0, N )
t,x,v = verlet( dt, t0, x0, v0, N )

#Create figure and axis
fig1 = plt.figure(figsize=(10, 8))

#Plot x vs t
plt.plot( te,xe,'-',label = "Euler" , color = [1,0,0] )
plt.plot( t,x,'-',label = "Verlet", color = [0,0,1] )
plt.xlabel( "t", size=fsize );
plt.ylabel( "x", size=fsize );
plt.xticks( fontsize = fsize )
plt.yticks( fontsize = fsize )
plt.title( title_str, size=fsize )
plt.legend( fontsize = fsize,loc = "upper right" )
plt.xlim( t0, tmax )
plt.ylim( -xmax, xmax )
plt.grid(True)

#Export PNG file
fig1.savefig('x vs t.png', dpi=300)

#Create figure and axis
fig2 = plt.figure(figsize=(10, 8))

#Plot x vs t
plt.plot( te,ve,'-',label = "Euler" , color = [1,0,0] )
plt.plot( t,v,'-',label = "Verlet", color = [0,0,1] )
plt.xlabel( "t", size=fsize );
plt.ylabel( "v", size=fsize );
plt.xticks( fontsize = fsize )
plt.yticks( fontsize = fsize )
plt.title( title_str, size=fsize )
plt.legend( fontsize = fsize,loc = "upper right" )
plt.xlim( t0, tmax )
plt.ylim( -vmax, vmax )
plt.grid(True)

#Export PNG file
fig2.savefig('v vs t.png', dpi=300)

#Wait for user input to close terminal window
input("\n Press return key to close ...")

#End of code
